import numpy as np

def hermitegaussian(coeffs,x,sigma):
    xhat = (x/sigma)
    herms = np.polynomial.hermite.Hermite(coeffs)
    return herms(xhat) * np.exp(-xhat**2)

def continuous_convolve(kernels,obj):
    out = np.empty(obj.shape)
    for i in range(kernels.shape[0]):
        out[jj] = np.dot(obj[max(0,jj-centering):min(size,jj+n-centering)], kernels[i,max(0,centering-jj):min(n,size-jj+centering)])
    return out

def convolve_hermites(f_in,coeffs,center_kw,sigma,sigma_range,spacing):
    x = np.arange(-sigma_range * max(sigma),sigma_range * max(sigma),step=spacing)

    if center_kw == 'centered':
        centering = int(x.shape[0]/2)
    elif center_kw == 'right':
        centering = 0
    elif center_kw == 'left':
        centering = x.shape[0]-1
    else:
        print('setting lsf centering to middle')
        centering = int(x.shape[0]/2)

    f_out = np.empty(f_in.shape)
    size = f_in.shape[0]
    n    = x.shape[0]

    for jj in range(f_out.shape[0]):
        kernel = hermitegaussian(coeffs[jj,:],x,sigma[jj])
        # L1 normalize the kernel so the total flux is conserved
        kernel /= np.sum(kernel)
        f_out[jj] = np.dot(f_in[max(0,jj-centering):min(size,jj+n-centering)]\
                                    ,kernel[max(0,centering-jj):min(n,size-jj+centering)])
    return f_out



import numpy as np
import astropy.units as u
import astropy.time as at
import astropy.coordinates as coord
import scipy.interpolate as interp
import scipy.ndimage as img
import scipy.sparse
import numpy.random as random

def dict_from_h5(hf,data):

    import h5py
    for key in hf.keys():
        if key == 'obs_times':
            try:
                print(hf[key])
                data[key] = at.Time(np.array(hf[key]).tolist(),format='isot')
            except TypeError:
                pass
        elif key == 'loc':
            # print(hf[key]['value'][()],hf[key]['unit'][()][0])
            data[key] = coord.EarthLocation.from_geocentric(hf['loc']['value'][()][0] * u.Unit(hf['loc']['unit'][()][0]) \
                            ,hf['loc']['value'][()][1] * u.Unit(hf['loc']['unit'][()][0]) \
                            ,hf['loc']['value'][()][2] * u.Unit(hf['loc']['unit'][()][0]))
        elif key == 'target':
            data[key] = coord.SkyCoord(hf['target']['ra'][0] * u.deg,hf['target']['dec'][0] * u.deg)
        elif key == 'value':
            return np.array(hf['value']) * u.Unit(hf['unit'][0])
        elif isinstance(hf[key], h5py.Group):
            data[key] = {}
            data[key] = dict_from_h5(hf[key],data[key])
        elif len(hf[key].shape) == 0:
            data[key] = hf[key]
        else:
            data[key] = np.array(hf[key])
    return data

def from_h5(filename):
    import h5py
    data = {}
    with h5py.File(filename,'r') as file:
        data = dict_from_h5(file,data)

    return DetectorData(data)

def convert_xy(x,y,yerr=None,units=u.Angstrom):
    outerr = None
    if yerr is not None:
        outerr = yerr / y
    return np.log(x.to(units).value), np.log(y), outerr

def save_dict_as_h5(hf,data):
    import h5py
    for key in data.keys():
        if isinstance(data[key],u.Quantity):
            print('quantity')
            group = hf.create_group(key)
            group.create_dataset('value',data=data[key].value)
            dt = h5py.special_dtype(vlen=str)
            unt = np.array([str(data[key].unit)],dtype=dt)
            group.create_dataset('unit',data=unt)
        elif isinstance(data[key], np.ndarray):
            if data[key].dtype == at.Time:
                print('saving time')
                dt = h5py.special_dtype(vlen=str)
                times = np.array([x.strftime('%Y-%m-%dT%H:%M:%S.%f%z') for x in data[key]],dtype=dt)
                hf.create_dataset(key,data=times)
            else:
                hf.create_dataset(key,data=data[key])
        # except AttributeError:
        elif isinstance(data[key], dict):
            group = hf.create_group(key)
            save_dict_as_h5(group,data[key])
        elif isinstance(data[key], coord.EarthLocation):
            print('saving location...{}'.format(data[key].geocentric))
            hf.create_dataset(key,data[key].geocentric)
        elif isinstance(data[key],coord.SkyCoord):
            print('saving target...{} {}'.format(data[key].ra,data[key].dec))
            group = hf.create_group(key)
            group.create_dataset('ra',data=[data[key].ra.to(u.deg).value])
            group.create_dataset('dec',data=[data[key].dec.to(u.deg).value])
        else:
            print(key, ' saving as string')
            dt = h5py.special_dtype(vlen=str)
            arr = np.array([str(data[key])],dtype=dt)
            hf.create_dataset(key,data=arr)

def from_pickle(filename):
    import pickle
    with open(filename, 'rb') as input:  # Overwrites any existing file.
        model = pickle.load(input)
        return model

data_plot_settings = {'marker':'.','color':'black','alpha':0.9,'label':'Data'}
interpolated_settings = {'marker':'.','alpha':0.3,'label':'Interpolated','markersize':3}
truth_settings = {'marker':'o','alpha':0.4,'label':'Truth','markersize':4}

gas_settings = {'color':'green'}
star_settings = {'color':'red'}
tellurics_settings = {'color':'blue'}
lsf_settings = {'color':'pink'}

def print_keys(data,extra=''):
    for key in data.keys():
        print(extra + '{}: {}'.format(key, type(data[key])))
        if isinstance(data[key],dict):
            tab = '--'
            extra += tab
            print_keys(data[key],extra)
            extra = extra[:-len(tab)]

class DetectorData:
    def __init__(self,data={}):
        self.data = data

    def to_h5(self,filename):
        import h5py
        hf = h5py.File(filename,"w")
        save_dict_as_h5(hf,self.data)
        hf.close()

    def to_pickle(self,filename):
        import pickle
        with open(filename, 'wb') as output:  # Overwrites any existing file.
            pickle.dump(self, output, pickle.HIGHEST_PROTOCOL)

    def to_fits(self,filename):
        import astropy.io.fits as fits

        hdul = fits.HDUList([])
        for key in self.keys():
            print(key)
            for subkey in self[key].keys():
                print('\t',subkey)
                image_hdu = fits.ImageHDU(self[key][subkey])
                hdul.append(image_hdu)

        hdul.writeto(filename)

    def keys(self):
        return self.data.keys()

    def __getitem__(self,key):
        return self.data[key]

    def __setitem__(self,key,value):
        self.data[key] = value

    def plot_flux(self,ax,i,flux_keys,wave_keys,pargs=[],ferr_keys=None,xy='',units=u.Angstrom,normalize=None,nargs=[]):
        y_data = self
        for key in flux_keys:
            y_data = y_data[key]
        if len(y_data.shape) == 1:
            y = y_data
        else:
            y = y_data[i,:]

        x_data = self
        for key in wave_keys:
            x_data = x_data[key]
        if len(x_data.shape) == 1:
            x = x_data
        else:
            x = x_data[i,:]

        yerr = 0.0
        if ferr_keys is not None:
            err_data = self
            for key in ferr_keys:
                err_data = err_data[key]
            yerr = err_data[i,:]

        if normalize is not None:
            y, yerr = normalize(y,yerr,*nargs)
        if 'x' in xy:
            x = np.log(x.to(u.Angstrom).value)
            # x, y, _ = convert_xy(x, y, None, units=units)
        else:
            x = x.value
        if 'y' in xy:
            y, yerr = np.log(y), yerr/y
        if ferr_keys is None:
            ax.plot(x, y, **pargs)
        else:
            ax.errorbar(x, y, yerr,**pargs)
        return ax

    def plot_data(self,ax,i,xy='',units=u.Angstrom,normalize=None,nargs=[]):
        self.plot_flux(ax,i,['data','flux'],['data','wave'],ferr_keys=['data','ferr'],pargs=data_plot_settings,xy=xy,units=units,normalize=normalize,nargs=nargs)
        return ax

    def plot_theory(self,ax,i,xy='',units=u.Angstrom,normalize=None,nargs=[]):
        self.plot_flux(ax,i,['theory','interpolated','total','flux'],['theory','interpolated','total','wave'],xy=xy,pargs={'color':'gray',**interpolated_settings},units=units,normalize=normalize,nargs=nargs)
        return ax

    def plot_lsf(self,ax,i,xy='',units=u.Angstrom,normalize=None,nargs=[]):
        self.plot_flux(ax,i,['theory','lsf','flux'],['theory','interpolated','total','wave'],xy=xy,pargs={'color':'pink',**interpolated_settings},units=units,normalize=normalize,nargs=nargs)
        return ax

    def plot_star(self,ax,i,xy='',units=u.Angstrom,normalize=None,nargs=[]):
        self.plot_flux(ax,i,['theory','interpolated','star','flux'],['theory','interpolated','total','wave'],pargs={**star_settings,**interpolated_settings},xy=xy,units=units,normalize=normalize,nargs=nargs)
        return ax

    def plot_gas(self,ax,i,xy='',units=u.Angstrom,normalize=None,nargs=[]):
        self.plot_flux(ax,i,['theory','interpolated','gascell','flux'],['theory','interpolated','total','wave'],xy=xy,units=units,pargs={**gas_settings,**interpolated_settings},normalize=normalize,nargs=nargs)
        return ax

    def plot_tellurics(self,ax,i,xy='',units=u.Angstrom,normalize=None,nargs=[]):
        self.plot_flux(ax,i,['theory','interpolated','tellurics','flux'],['theory','interpolated','total','wave'],xy=xy,pargs={**tellurics_settings,**interpolated_settings},normalize=normalize,nargs=nargs)
        return ax

    def plot_rvs(self,ax,units=u.km/u.s):
        now = at.Time.now()
        time = np.array([(x - at.Time.now()).value for x in self['data']['obs_times']],dtype=float)
        rvs = self['data']['rvs'].to(units).value
        ax.plot(time,rvs,'.k')
        return ax

    def plot_rvs_minus_bcs(self,ax,units=u.km/u.s):
        now = at.Time.now()
        bcs = simulacra.star.get_berv(self['data']['obs_times'],self['parameters']['obs'],
                                        self['parameters']['ra'],self['parameters']['dec'],
                                        self['parameters']['velocity_drift'])
        ax.plot(self['data']['times'] - now,(self['parameters']['rvs'] - bcs).to(units).value,'.k')
        return ax

    def plot_rvs_minus_bcs_mod_period(self,ax,units=u.km/u.s):
        now = at.Time.now()
        bcs = simulacra.star.get_berv(self['data']['obs_times'],self['parameters']['obs'],
                                        self['parameters']['ra'],self['parameters']['dec'],
                                        self['parameters']['velocity_drift'])
        ax.plot((self['data']['times'] - now) % self['parameters']['period'],(self['parameters']['rvs'] - bcs).to(units).value,'.k')
        return ax



import numpy as np
import numpy.random as random

import sys
import h5py

import scipy.ndimage as img
import scipy.interpolate as interp
import scipy.io

import astropy.units as u
import astropy.constants as const
import astropy.table as at
import astropy.io

import data.tellurics.skycalc.skycalc as skycalc
import data.tellurics.skycalc.skycalc_cli as sky_cli

import json
import requests
import io

    # WobbleSim/wobblesim.py
    # WobbleSim/cli/main.py


def main(low_resolution,s2n,epoches,vp,epsilon,gamma,w,stellarname_wave,stellarname_flux,skycalcname,skycalcalma,gascellname,a):
    generate_data = False

    # Read In Files And Simulate
    #################################################
    # lambda is in Angstroms here
    lamb_min = 5000
    lamb_max = 6300
    xmin   = np.log(lamb_min)
    xmax   = np.log(lamb_max)
    deltas = sample_deltas(epoches,vel_width=vp*u.km/u.s)
    flux_stellar, lamb_stellar = read_in_stellar(stellarname_wave,
                                                 stellarname_flux)

    # lambda here is in nanometers
    trans_tellurics, lamb_tellurics, airmass = simulate_tellurics(inputFilename=skycalcname,
                                                                    almFilename=skycalcalma,
                                                                    epoches=epoches)

    # lambda is in nanometers here as well
    trans_gas, lamb_gas = read_in_gas_cell(filename=gascellname)

    lamb_tellurics *= u.nm
    lamb_gas       *= u.Angstrom
    lamb_stellar   *= u.Angstrom

    x_s = np.log(lamb_stellar/u.Angstrom)
    x_t = np.log(lamb_tellurics/u.Angstrom)
    x_g = np.log(lamb_gas/u.Angstrom)

    temp = [get_median_difference(x) for x in [x_s,x_t[0],x_g]]
    median_diff = min(temp)

    xs  = np.arange(np.log(lamb_min),np.log(lamb_max),step=median_diff)
    f_s = np.empty((epoches,xs.shape[0]))
    f_t = np.empty((epoches,xs.shape[0]))
    f_g = interpolate(xs,x_g,trans_gas)

    f_theory = np.empty((epoches,xs.shape[0]))
    for i in range(epoches):
        f_s[i,:]   = interpolate(xs + deltas[i],x_s,     flux_stellar)
        f_t[i,:]   = interpolate(xs,x_t[i,:],trans_tellurics[i])
        f_theory[i,:] = f_s[i,:] * f_t[i,:] * f_g

    # now take lambda grids from all of these and make a new one with
    # spacing equal to the minimum median spacing of the above grids
    # then using lanczos 5 interpolation interpolate all values onto new grids
    # then multiply element wise for combined theoretical spectrum

    if generate_data:
        # Initialize constants
        #################################################
        high_spacing = spacing_from_res(high_resolution)
        xs           = np.arange(xmin,xmax,step=high_spacing)
        line_width   = spacing_from_res(low_resolution)

        # Generate Theoretical Y values
        #################################################
        y_star,deltas  = generate_stellar(  sn,epoches,xs,line_width,ymin,ymax,vp*u.km/u.s)
        y_tell,airmass = generate_tellurics(tn,epoches,xs,line_width,ymin,ymax)
        y_gas ,mu_g    = generate_gas_cell( gn,epoches,xs,line_width,ymin,ymax)

        y_sum = y_star + y_tell + y_gas
        f_sum = np.exp(y_sum)

    # Convolve with Telescope PSF
    ################################################
    lsf = mean_lsf(low_resolution,median_diff,sigma_range=5.0)
    f_lsf = np.empty(f_theory.shape)
    for iii in range(f_theory.shape[0]):
        f_lsf[iii,:] = img.convolve(f_theory[iii,:],lsf)
        # np.apply_along_axis(dummy,1,f_theory,lsf) # convolve just tell star and gas

    # Generate dataset grid & jitter & stretch
    ##################################################
    x   = np.arange(xmin,xmax,step=spacing_from_res(low_resolution))
    nlr = x.shape[0]
    x_hat, m    = stretch(x,epoches,epsilon)
    x_hat, delt = jitter(x,epoches,w)

    # Interpolate Spline and Add Noise
    ##################################################
    s2n_grid  = get_s2n(x_hat.shape,s2n)
    f_exp     = np.empty(x_hat.shape)
    f_readout = np.empty(x_hat.shape)
    # noise = np.empty(x_hat.shape)
    for i in range(f_exp.shape[0]):
        f_exp[i,:] = lanczos_interpolation(x_hat[i,:],xs,f_lsf[i,:],dx=median_diff,a=a)
        for j in range(f_exp.shape[1]):
            f_readout[i,j] = f_exp[i,j] * random.normal(1,1./s2n_grid[i,j])

    # Get Error Bars
    ###################################################
    ferr_out = generate_errors(f_readout,s2n_grid,gamma)
    lmb_out  = np.exp(x)

    # Pack Output into Dictionary
    ###################################################
    out = {"wavelength_sample":lmb_out,
            "flux":f_readout,
            "flux_error":ferr_out,
            "wavelength_theory":np.exp(xs),
            "flux_tellurics":f_t,
            "flux_stellar":f_s,
            "flux_gas":f_g,
            "flux_lsf":f_lsf,
            "del":delt,
            "m":m,
            "airmass":airmass,
            "delta":deltas}

    return out

def mean_lsf(low_resolution,spacing,sigma_range=5.0):
    lsf = np.arange(-sigma_range/low_resolution,sigma_range/low_resolution,step=spacing)
    lsf = gauss_func(lsf,mu=0.0,sigma=1.0/low_resolution)
    lsf /= np.linalg.norm(lsf,ord=1)
    return lsf

def lanczos_interpolation(x,xs,ys,dx,a=4):
    x0 = xs[0]
    y = np.zeros(x.shape)
    for i,x_value in enumerate(x):
        # which is basically the same as sample=x[j-a+1] to x[j+a]
        # where j in this case is the nearest index xs_j to x_value
#         print("value: ", x_value)
#         print("closest: ",xs[int((x_value-x0)//dx)])
#         print,x_value)
        sample_min,sample_max = max(0,abs(x_value-x0)//dx - a + 1),min(xs.shape[0],abs(x_value-x0)//dx + a)

        samples = np.arange(sample_min,sample_max,dtype=int)
#         print(sample_min,sample_max)
        for sample in samples:
            y[i] += ys[sample] * lanczos_kernel((x_value - xs[sample])/dx,a)
    return y

def lanczos_kernel(x,a):
    if x == 0:
        return 1
    if x > -a and x < a:
        return a*np.sin(np.pi*u.radian*x) * np.sin(np.pi*u.radian*x/a)/(np.pi**2 * x**2)
    return 0

def same_dist_elems(arr):
    diff = arr[1] - arr[0]
    for x in range(1, len(arr) - 1):
        if arr[x + 1] - arr[x] != diff:
            return False
    return True

def read_in_tellurics(filename):
    hdu = astropy.io.fits.open(filename)
    print(hdu)
    prim = hdu['PRIMARY'].header.keys
    print(prim)
    sys.exit()
    prim = at.Table.read(hdu['PRIMARY'])
    print(prim.info())
    tbl  = at.Table.read(hdu[1])

    sys.exit()
    trans_grid = np.array(tbl['trans'].data)
    lamb_grid  = np.array(tbl['lam'].data)
    airmass = np.array(prim['airmass'].data)

    trans_grid = np.expand_dims(trans_grid,0)
    lamb_grid  = np.expand_dims(lamb_grid,0)
    return trans_grid, lamb_grid, airmass

def get_s2n(shape,constant):
    return np.ones(shape) * constant

def average_difference(x):
    return np.mean([t - s for s, t in zip(x, x[1:])])

def jitter(x,epoches,w=1.0):
    out = x
    if len(out.shape) == 1:
        out = np.expand_dims(out,axis=0)
        out = np.repeat(out,repeats=epoches,axis=0)

    width = average_difference(out[0,:])
    jitter = (2*random.rand(epoches) - 1) * width * w
    for i,delt in enumerate(jitter):
        out[i,:] += delt
    return out,jitter

def stretch(x,epoches,epsilon=0.01):
    if len(x.shape) == 1:
        x = np.expand_dims(x,axis=0)
        x = np.repeat(x,repeats=epoches,axis=0)
    m = (epsilon * (2*random.rand(epoches) - 1)) + 1
    for i,ms in enumerate(m):
        x[i,:] *= ms
    return x,m

def gauss_func(x,mu,sigma):
    return np.exp((-1/2)*(x - mu)**2/(sigma**2))

def generate_noise(epoches,size,scale=0.01):
    return random.normal(scale=scale,size=(epoches,size))

def spacing_from_res(R):
    return np.log(1+1/R)

def generate_stellar(n_lines,epoches,x,line_width,y_min=0.0,y_max=0.7,vel_width=30*u.km/u.s):
    deltas  = np.array(shifts((2*random.rand(epoches)-1)*vel_width))
    mus     = (np.max(x) - np.min(x))*random.rand(n_lines) + np.min(x)
    heights = (y_max - y_min) * random.rand(n_lines) + y_min

    y = np.zeros((epoches,x.shape[0]))
    for i,delta in enumerate(deltas):
        for j in range(n_lines):
            y[i,:] -= heights[j] * gauss_func(x + delta,mus[j],line_width)
    return y, deltas

def generate_tellurics(n_lines,epoches,x,line_width,y_min=0.0,y_max=0.7):
    airmass = random.rand(epoches)
    mus     = (np.max(x) - np.min(x))*random.rand(n_lines) + np.min(x)
    heights = (y_max - y_min) * random.rand(n_lines) + y_min

    y = np.zeros((epoches,x.shape[0]))
    for i in range(epoches):
        for j in range(n_lines):
            y[i,:] -= airmass[i] * heights[j] * gauss_func(x,mus[j],line_width)
    return y, airmass

def generate_gas_cell(n_lines,epoches,x,line_width,y_min=0.0,y_max=0.7):
    mus     = (np.max(x) - np.min(x))*random.rand(n_lines) + np.min(x)
    heights = (y_max - y_min) * random.rand(n_lines) + y_min

    y = np.zeros((epoches,x.shape[0]))
    for i in range(epoches):
        for j in range(n_lines):
            y[i,:] -= heights[j] * gauss_func(x,mus[j],line_width)
    return y, mus

def generate_errors(f,s2n,gamma=1.0):
    xs,ys = np.where(f < 0)
    for x,y in zip(xs,ys):
        f[x,y] = 0
    f_err = np.empty(f.shape)
    for i in range(f_err.shape[0]):
        for j in range(f_err.shape[1]):
            error = random.normal(scale=f[i,j]/s2n[i,j] * gamma)
            # print(error)
            f_err[i,j] = error
    return f_err



import numpy as np
import astropy.units as u
import astropy.constants as const
import astropy.time as at
import scipy.interpolate as interp
import scipy.ndimage as img
import scipy.sparse
import numpy.random as random
import logging

from simulacra.dataset import DetectorData
import simulacra.lanczos
import simulacra.convolve


from itertools import repeat
from multiprocessing import Pool

def dict_of_attr(data,obj):
    obj_list = [a for a in dir(obj) if not a.startswith('__')]
    for ele in obj_list:
        try:
            data[ele] = getattr(obj, ele)
        except AttributeError:
            pass
    return data

def get_median_difference(x):

    return np.median([t - s for s, t in zip(x, x[1:])])

def spacing_from_res(R):
    return np.log(1+1/R)

def jitter(x,epoches,w=1.0):
    out = x
    if len(out.shape) == 1:
        out = np.expand_dims(out,axis=0)
        out = np.repeat(out,repeats=epoches,axis=0)

    width = average_difference(out[0,:])
    jitter = (2*random.rand(epoches) - 1) * width * w
    for i,delt in enumerate(jitter):
        out[i,:] += delt
    return out,jitter

def stretch(x,epoches,epsilon=0.01):
    if len(x.shape) == 1:
        x = np.expand_dims(x,axis=0)
        x = np.repeat(x,repeats=epoches,axis=0)
    m = (epsilon * (2*random.rand(epoches) - 1)) + 1
    for i,ms in enumerate(m):
        x[i,:] *= ms
    return x,m

def interpolate(x,xs,ys):
    spline = interp.CubicSpline(xs,ys)
    return spline(x)

def average_difference(x):
    return np.mean([t - s for s, t in zip(x, x[1:])])

def generate_errors(f,snr):
    xs,ys = np.where(f < 0)
    for x,y in zip(xs,ys):
        f[x,y] = 0
    f_err = np.empty(f.shape)
    for i in range(f_err.shape[0]):
        for j in range(f_err.shape[1]):
            f_err[i,j] = f[i,j]/snr[i,j]
    return f_err

from numba import vectorize, float64
@vectorize([float64(float64, float64)])
def add_noise_v(f, snr):
    return f + random.normal(0.0,f/snr)

@vectorize([float64(float64, float64)])
def generate_errors_v(f, snr):
    return f / snr

def add_noise(f_exp,snr_grid):
    f_readout = np.empty(f_exp.shape)
    for i in range(f_exp.shape[0]):
        print('snr {}: {}'.format(i,np.median(snr_grid[i,:])))
        for j in range(f_exp.shape[1]):
            f_readout[i,j] = f_exp[i,j] + random.normal(0.0,f_exp[i,j]/snr_grid[i,j])
    return f_readout

# def signal_to_noise_ratio(detector,flux,exp_times):
#     xs,ys = np.where(flux < 0)
#     for x,y in zip(xs,ys):
#         flux[x,y] = 0
#
#     snr = np.empty(flux.shape)
#     for i in range(snr.shape[0]):
#         for j in range(snr.shape[1]):
#             # if j % 500 == 0:
#             #     print(detector.read_noise[j], (detector.dark_current[j] * exp_times[i]).to(1), detector.ccd_eff[j] * flux[i,j])
#             # implicitly flux is already dependent on exposure time
#             snr[i,j] = detector.ccd_eff[j] * flux[i,j] \
#             / (detector.gamma * np.sqrt(detector.read_noise[j] + detector.dark_current[j] * exp_times[i] + detector.ccd_eff[j] * flux[i,j]))
#     return snr

def interpolate_mask(xs,mask_the,x_hat):
    return np.array([interp.interp1d(xs,mask_the[i,:].astype(float),kind='nearest')(x_hat[i,:]) for i in range(x_hat.shape[0])]).astype(bool)

def check_type(value,type):
    if isinstance(value,u.Quantity):
        if value.unit.physical_type in u.get_physical_type(type.unit):
            pass
        else:
            logging.error('must be in a units of {}.\nor will be assumed to be in units of photons per second.'.format(type))
    else:
        value *= type
    return value

def check_shape(value,shape):
    if hasattr(value,'shape'):
        if len(value.shape) == 0:
            return value * np.ones(shape)
        elif value.shape == shape:
            return value
    else:
        return value * np.ones(shape)


class Detector:
    def __init__(self,stellar_model,resolution,loc,area,wave_grid,dark_current,read_noise,ccd_eff,through_put=0.2,wave_padding=5*u.Angstrom,epsilon=0.0,gamma=1.0,w=0.0,a=4):
        '''Detector model that simulates spectra from star given resolution...

        Detector takes on a given theoretical `stellar_model`, `resolution`, with
        a constant signal to noise ratio, `snr`. Then call simulate to generate a given number
        of epoches of spectral data.

        Parameters
        ----------
        stellar_model: TheoryModel with a generate_data method that returns flux, wave, and deltas
        resolution: the resolution of the detector being simulated
        epsilon: a small that is used to pull the random stretchs for the wavelength grid, m ~ uniform(1-epsilon,1+epsilon)
        snr: the constant signal to noise ratio
        gamma: a constant that random effects the errorbars of flux \sigma_f ~ normal(mu=0.0,sigma=gamma * f_i,j/snr_i,j)
        w: a constant that random affects the jitter to the wavelength grid, del ~ uniform(-w*pxl_width, w*pxl_width)
        a: the number of kernel functions used at the lanczos interpolation step

        Returns
        -------
        data: DetectorData type that contains all parameters generate
        '''
        # Simulator Models
        self.stellar_model = stellar_model
        self.transmission_models = []

        # Randomized Parameters
        self.epsilon = epsilon
        self.gamma   = gamma
        self.w       = w
        self.a       = a

        # Simulation Parameters
        self._lambmin = 0.0 * u.nm
        self._lambmax = 100000 * u.nm

        # Detector Properties
        self.wave_grid    = wave_grid
        self.wave_padding = wave_padding
        self.dark_current = dark_current
        self.read_noise   = read_noise
        self.ccd_eff      = ccd_eff
        self.through_put  = through_put
        self.area = area
        self.loc  = loc
        # LSF properties
        self.sigma_range   = 5.0
        self.resolution   = resolution
        self.lsf_const_coeffs = [1.0]
        # self.sigma      = 1.0/resolution

        self.lsf_centering = 'centered'

        self.transmission_cutoff = 10.

    def res(self,wavelength):
        if isinstance(self._resolution, float):
            return self._resolution * np.ones(wavelength.shape)
        elif hasattr(self._resolution, '__call__'):
            return self._resolution(wavelength)
        else:
            logging.error('resolution has not been set.')
            return 0

    def resolution():
        doc = "The resolution property."
        def fset(self, value):
            if isinstance(value,float):
                self._resolution = value
            elif hasattr(value, '__call__'):
                self._resolution = value
            else:
                logging.error('resolution grid needs to be a single value \
                                or a callable that takes wavelength as input')
        def fdel(self):
            logging.warn('overwriting resolution')
            del self._resolution
        return locals()
    resolution = property(**resolution())

    def transmission():
        doc = "The transmission property."
        def fget(self):
            return self._transmission
        def fset(self, value):
            self._transmission = value
        def fdel(self):
            del self._transmission
        return locals()
    transmission = property(**transmission())

    def lambmin():
        doc = "The lambmin property."
        def fget(self):
            return np.min(self.wave_grid) - self.wave_padding
        return locals()
    lambmin = property(**lambmin())

    def lambmax():
        doc = "The lambmax property."
        def fget(self):
            return np.max(self.wave_grid) + self.wave_padding
        return locals()
    lambmax = property(**lambmax())

    def wave_grid():
        doc = "The wave_grid property."
        def fget(self):
            return self._wave_grid
        def fset(self, new_grid):
            minimum = self.checkmin()
            maximum = self.checkmax()
            if minimum >= maximum:
                logging.error('no overlap between selected wave grids\nmodel cannot be added')
                self.transmission_models.pop()
                return
            self._wave_grid = new_grid[np.multiply(new_grid <= maximum, new_grid >= minimum,dtype=bool)]
            if np.min(new_grid) < minimum:
                print("wave_grid min -> {}".format(minimum))
            if np.max(new_grid) > maximum:
                print("wave_grid min -> {}".format(maximum))
        return locals()
    wave_grid = property(**wave_grid())

    def wave_difference():
        doc = "the wave_difference property"
        def fget(self):
            try:
                return self._wave_difference
            except AttributeError:
                diff = self.wave_grid[1:] - self.wave_grid[:-1]
                return np.concatenate(([np.mean(diff)],diff))
        def fset(self,value):
            if isinstance(value, np.ndarray):
                if value.shape == self.wave_grid.shape:
                    self._wave_difference = value
                else:
                    logging.error('difference array must have the same shape as wave grid {}.'.format(value.shape,self.wave_grid.shape))
            else:
                self._wave_difference = value * np.ones(self.wave_grid.shape)
        def fdel(self):
            self._wave_difference = None
        return locals()
    wave_difference = property(**wave_difference())

    def ccd_eff():
        doc = "The ccd_eff property."
        def fget(self):
            return self._ccd_eff
        def fset(self, value):
            value = check_shape(value,self.wave_grid.shape)
            self._ccd_eff = value
        def fdel(self):
            del self._ccd_eff
        return locals()
    ccd_eff = property(**ccd_eff())

    def dark_current():
        doc = "The dark_current property."
        def fget(self):
            return self._dark_current
        def fset(self, value):
            value = check_type(value,1/u.s)
            value = check_shape(value,self.wave_grid.shape)
            self._dark_current = value
        def fdel(self):
            del self._dark_current
        return locals()
    dark_current = property(**dark_current())

    def read_noise():
        doc = "The read_noise property."
        def fget(self):
            return self._read_noise
        def fset(self, value):
            value = check_shape(value,self.wave_grid.shape)
            self._read_noise = value
        def fdel(self):
            del self._read_noise
        return locals()
    read_noise = property(**read_noise())

    def add_model(self,model):
        self.transmission_models.append(model)
        # after adding model to list
        # reset the wave_grid so that it can be truncated
        self.wave_grid = self.wave_grid

    def checkmin(self):
        minimums = [self.stellar_model.lambmin]
        for model in self.transmission_models:
            minimums += [model.lambmin]
        # print(minimums)
        return max(minimums)

    def checkmax(self):
        maximums = [self.stellar_model.lambmax]
        for model in self.transmission_models:
            maximums += [model.lambmax]
        # print(maximums)
        return min(maximums)

    def simulate(self,obs_times,exp_times):
        data = DetectorData()
        data['data'] = {}
        data['data']['obs_times'] = obs_times
        data['data']['exp_times'] = exp_times
        data['data']['epoches']   = obs_times.shape[0]
        epoches = obs_times.shape[0]

        # Generate Stellar Spectra
        ###################################################
        data['theory'] = {}
        data['theory']['star'] = {}
        flux_stellar, wave_stellar, deltas, rvs = self.stellar_model.generate_spectra(self,obs_times,exp_times)

        data['data']['rvs'], data['theory']['star']['deltas'] = rvs, deltas
        data['theory']['star']['flux'], data['theory']['star']['wave'] = flux_stellar, wave_stellar
        differences = [get_median_difference(np.log(wave_stellar.to(u.Angstrom).value))]

        # Generate Transmission
        ###################################################
        print('generating spectra...')
        trans_flux, trans_wave = [], []
        data['theory']['interpolated'] = {}
        for model in self.transmission_models:
            data['theory']['interpolated'][model._name] = {}
            data['theory'][model._name] = {}
            flux, wave = model.generate_transmission(self.stellar_model,self,obs_times,exp_times)
            trans_flux.append(flux), trans_wave.append(wave)
            differences += [get_median_difference(np.log(wave[iii][:].to(u.Angstrom).value)) for iii in range(len(wave))]
            data['theory'][model._name]['flux'],data['theory'][model._name]['wave'] = flux, wave
            print(model, differences)
        new_step_size = min(differences)

        # Interpolate all models and combine onto detector
        # PARALLELIZE
        ##################################################################
        data['theory']['interpolated']['star'] = {}
        xs = np.arange(np.log(self.lambmin.to(u.Angstrom).value),np.log(self.lambmax.to(u.Angstrom).value),step=new_step_size)
        print('interpolating spline...')
        stellar_arr = np.empty((epoches,xs.shape[0]))
        trans_arrs  = np.empty((len(self.transmission_models),epoches,xs.shape[0]))
        for i in range(epoches):
            print(i)
            stellar_arr[i,:] = self.interpolate_grid(xs + deltas[i],np.log(wave_stellar.to(u.Angstrom).value),flux_stellar[i,:].to(u.erg/u.s/u.cm**3).value)
            for j,model in enumerate(self.transmission_models):
                trans_arrs[j,i,:] = self.interpolate_grid(xs,np.log(trans_wave[j][i][:].to(u.Angstrom).value),trans_flux[j][i][:])
        print('combining grids...')
        data['theory']['interpolated']['star']['flux'] = stellar_arr
        fs        = stellar_arr.copy()
        flux_unit = flux_stellar.unit
        mask_the  = np.zeros(fs.shape,dtype=bool)
        for j,model in enumerate(self.transmission_models):
            # print("fs: ", fs.shape)
            fs *= trans_arrs[j,:,:]
            mask_the = (trans_arrs[j,:,:] > self.transmission_cutoff) | mask_the
            data['theory']['interpolated'][model._name]['flux'] = trans_arrs[j,:,:]
        data['theory']['interpolated']['total'] = {}
        data['theory']['interpolated']['total']['flux'] = fs
        data['theory']['interpolated']['total']['wave'] = np.exp(xs) * u.Angstrom
        data['theory']['interpolated']['total']['mask'] = mask_the

        # Convolving using Hermite Coeffs
        #################################################
        # should be an array that can vary over pixel j or hermite m
        print('convolving...')
        f_lsf = self.convolve(xs,fs,new_step_size)

        data['theory']['lsf'] = {}
        data['theory']['lsf']['flux'] = f_lsf

        # Generate transform wavelength grid using jitter & stretch
        ##################################################
        x                    = np.log(self.wave_grid.to(u.Angstrom).value)#np.arange(self.xmin,self.xmax,step=res_step_size)
        x_hat, wt_parameters = self.wave_transform(x,obs_times,exp_times)
        # x_hat, m    = stretch(x,epoches,self.epsilon)
        # x_hat, delt = jitter(x,epoches,self.w)
        data['parameters'] = {}
        data['parameters']['wavetransform'] = wt_parameters
        # data['parameters']['wavetransform']['m'] = m
        # data['parameters']['wavetransform']['delt'] = delt

        print('xs: {} {}\nxhat: {} {}'.format(np.exp(np.min(xs)),np.exp(np.max(xs)),np.exp(np.min(x_hat)),np.exp(np.max(x_hat))))
        data_mask = interpolate_mask(xs,mask_the,x_hat)
        data['data']['mask'] = data_mask

        # Interpolate using Lanczos and Add Noise
        ##################################################
        print('interpolating lanczos...')
        f_exp = self.interpolate_data(x_hat,xs,f_lsf)

        print('area: {}\t avg d lambda: {}\t avg lambda: {}\t avg exp times: {}'.format(self.area,np.mean(self.wave_difference),np.mean(self.wave_grid),np.mean(exp_times)))
        n_exp = self.energy_to_photon(f_exp * flux_unit, exp_times)
        # n_exp = self.through_put * (self.area/(const.hbar * const.c)*np.einsum('ij,j,j,i->ij',f_exp * flux_unit, self.wave_difference,self.wave_grid,exp_times)).to(1)
        for i in range(n_exp.shape[0]):
            print('{} n mean: {:3.2e}\t n median: {:3.2e}'.format(i,np.mean(n_exp[i,~data_mask[i,:]]),np.median(n_exp[i,~data_mask[i,:]])))

        print('generating true signal to noise ratios...')
        snr_grid = self.signal_to_noise_ratio(n_exp,exp_times)
        print('adding noise...')
        out_shape = snr_grid.shape
        n_readout = add_noise_v(n_exp.flatten().value,snr_grid.flatten()).reshape(out_shape)

        data['parameters']['true_snr'] = snr_grid
        data['data']['flux_expected'] = n_exp
        data['data']['flux'] = n_readout
        data['data']['wave'] = self.wave_grid

        # Get Error Bars
        ###################################################
        print('generating exp signal to noise ratios...')
        snr_readout = self.signal_to_noise_ratio(n_readout,exp_times)
        print('generating errors...')
        nerr_out = generate_errors_v(n_readout.flatten(),snr_readout.flatten()).reshape(out_shape)

        data['data']['snr_readout'] = snr_readout
        data['data']['ferr']        = nerr_out

        # Pack Parameters into Dictionary
        ###################################################
        data['parameters']['star'] = {}
        data['parameters']['star'] = dict_of_attr(data['parameters']['star'],self.stellar_model)

        data['parameters']['detector'] = {}
        data['parameters']['detector'] = dict_of_attr(data['parameters']['detector'],self)

        for model in self.transmission_models:
            data['parameters'][model._name] = {}
            data['parameters'][model._name] = dict_of_attr(data['parameters'][model._name],model)
        print('done.')
        return data

    def convolve(self,xs,fs,dx):
        self.lsf_coeffs = np.outer(np.ones((fs.shape[1],len(self.lsf_const_coeffs))), self.lsf_const_coeffs)
        class ConvolveIter:
            def __init__(obj,fs):
                obj.fs = fs
                obj._i =  0

            def __iter__(obj):
                # output = (fs[obj._i,:], self.lsf_coeffs,self.lsf_centering,self.sigma,self.sigma_range,new_step_size)
                obj._i = 0
                return obj

            def __next__(obj):

                if obj._i == fs.shape[0]:
                    raise StopIteration
                print(obj._i)
                output = (fs[obj._i,:], self.lsf_coeffs,\
                        self.lsf_centering,1./self.res(np.exp(xs) * u.Angstrom),\
                        self.sigma_range,dx)
                obj._i += 1
                return output

        with Pool() as pool:
            obj = ConvolveIter(fs)
            M = pool.starmap(simulacra.convolve.convolve_hermites, obj)
        f_lsf = np.asarray(M)
        return f_lsf

    def interpolate_grid(self,xs,x,f):
        spline = interp.CubicSpline(x,f)
        return spline(xs)

    def interpolate_data(self,xs,x,f):
        dx = average_difference(x)
        class LanczosIter:
            def __init__(obj):

                obj._i =  0

            def __iter__(obj):
                # output = (fs[obj._i,:], self.lsf_coeffs,self.lsf_centering,self.sigma,self.sigma_range,new_step_size)
                obj._i = 0
                return obj

            def __next__(obj):
                if obj._i == f.shape[0]:
                    raise StopIteration
                print(obj._i)
                output = (xs[obj._i,:],x,f[obj._i,:],dx,self.a)
                obj._i += 1
                return output

        with Pool() as pool:
            obj = LanczosIter()
            M = pool.starmap(simulacra.lanczos.lanczos_interpolation, obj)
        f_exp = np.asarray(M)
        return f_exp

    def wave_transform(self,x,obs_times,exp_times):
        epoches = obs_times.shape[0]
        x_hat, m    = stretch(x,epoches,self.epsilon)
        x_hat, delt = jitter(x,epoches,self.w)
        parameters = {'m':m,'delt':delt}
        return x_hat, parameters

    def signal_to_noise_ratio(self,flux,exp_times):
        xs,ys = np.where(flux < 0)
        for x,y in zip(xs,ys):
            flux[x,y] = 0

        snr = np.empty(flux.shape)
        for i in range(snr.shape[0]):
            for j in range(snr.shape[1]):
                # implicitly flux is already dependent on exposure time
                snr[i,j] = self.ccd_eff[j] * flux[i,j] \
                / (self.gamma * np.sqrt(self.read_noise[j] + self.dark_current[j] * exp_times[i] + self.ccd_eff[j] * flux[i,j]))
        return snr

    def energy_to_photon(self,flux,exp_times):
        n = self.through_put * (self.area/(const.hbar * const.c)*np.einsum('ij,j,j,i->ij',flux, self.wave_difference,self.wave_grid,exp_times)).to(1)
        return n



from simulacra.star import PhoenixModel
from simulacra.detector import Detector
from simulacra.tellurics import TelFitModel

import astropy.units as u

def run_expres(*args):
    # parser args into these constants and filename
    filename = 'out/somethingsomethingsometing.h5'
    logg = 1.0
    T = 4600
    z = -1.0
    alpha = 0.4
    epoches = 30
    resolution = 135000
    ra, dec = 1.0,10
    period = 300 * u.day
    amplitude = 2 * u.m/u.s
    # no jitter or stretch to wavelength grid
    epsilon = 0.0
    w = 0.0

    stellar_model   = PhoenixModel(alpha,z,T,alpha)
    tellurics_model = TelFitModel(1.5*u.micron,1.7*u.micron)

    detector = Detector(stellar_model,resolution,epsilon=epsilon,w=w)
    detector.add_model(tellurics_model)

    res,data = detector.simulate(epoches)
    data.to_h5(filename)
    # return data



# -*- coding: utf-8 -*-
"""skycalc Module"""

from __future__ import print_function
import sys
import json
from datetime import datetime
import requests


class AlmanacQuery:
    """AlmanacQuery for Querying the SkyCalc Almanac"""

    def __init__(self, indic):

        self.almdata = None
        self.almserver = 'https://etimecalret-002.eso.org'
        self.almurl = '/observing/etc/api/skycalc_almanac'

        # Left: users keyword (skycalc_cli),
        # Right: skycalc Almanac output keywords
        self.alm_parameters = {}
        self.alm_parameters['airmass'] = 'target_airmass'
        self.alm_parameters['msolflux'] = 'sun_aveflux'
        self.alm_parameters['moon_sun_sep'] = 'moon_sun_sep'
        self.alm_parameters['moon_target_sep'] = 'moon_target_sep'
        self.alm_parameters['moon_alt'] = 'moon_alt'
        self.alm_parameters['moon_earth_dist'] = 'moon_earth_dist'
        self.alm_parameters['ecl_lon'] = 'ecl_lon'
        self.alm_parameters['ecl_lat'] = 'ecl_lat'
        self.alm_parameters['observatory'] = 'observatory'

        self.almindic = {}
        # The Almanac needs:
        # coord_ra      : float [deg]
        # coord_dec     : float [deg]
        # input_type    : ut_time | local_civil_time | mjd
        # mjd           : float
        # coord_year    : int
        # coord_month   : int
        # coord_day     : int
        # coord_ut_hour : int
        # coord_ut_min  : int
        # coord_ut_sec  : float

        if 'date' in indic:
            self.almindic['input_type'] = 'ut_time'
            isotime = None
            try:
                isotime = datetime.strptime(indic['date'],
                                            '%Y-%m-%dT%H:%M:%S')
            except ValueError:
                print('Error: wrong date format for the Almanac.')
                raise
            self.almindic['coord_year'] = isotime.year
            self.almindic['coord_month'] = isotime.month
            self.almindic['coord_day'] = isotime.day
            self.almindic['coord_ut_hour'] = isotime.hour
            self.almindic['coord_ut_min'] = isotime.minute
            self.almindic['coord_ut_sec'] = isotime.second

        elif 'mjd' in indic:
            self.almindic['input_type'] = 'mjd'
            mjd = None
            try:
                mjd = float(indic['mjd'])
            except ValueError:
                print('Error: wrong mjd format for the Almanac.')
                raise
            self.almindic['mjd'] = mjd

        else:
            raise ValueError('Error: no date or mjd given for the Almanac')

        if 'ra' not in indic:
            raise ValueError('Error: ra coordinate not given for the Almanac.')

        if 'dec' not in indic:
            raise ValueError('Error: dec coordinate '
                             'not given for the Almanac.')

        ra = None
        try:
            ra = float(indic['ra'])
        except ValueError:
            print('Error: wrong ra format for the Almanac.')
            raise
        self.almindic['coord_ra'] = ra

        dec = None
        try:
            dec = float(indic['dec'])
        except ValueError:
            print('Error: wrong dec format for the Almanac.')
            raise
        self.almindic['coord_dec'] = dec

        if 'observatory' in indic:
            self.almindic['observatory'] = indic['observatory']

    def query(self):

        url = self.almserver + self.almurl

        rawdata = None
        try:
            response = requests.post(url, json.dumps(self.almindic))
            rawdata = response.text
        except requests.exceptions.RequestException as e:
            print('Error: Almanac query failed.')
            raise

        # Process rawdata
        jsondata = None
        try:
            jsondata = json.loads(rawdata)
            jsondata = jsondata['output']
        except (KeyError, ValueError):
            print('Error: invalid Almanac response.')
            raise

        # Find the relevant (key, value)
        almdata = {}
        for key, value in self.alm_parameters.items():
            subsection = 'nothing'
            prefix = value.split('_')[0]
            if prefix == 'sun' or prefix == 'moon' or prefix == 'target':
                subsection = prefix
            elif prefix == 'ecl':
                subsection = 'target'
            else:
                subsection = 'observation'
            try:
                almdata[key] = jsondata[subsection][value]
            except (KeyError, ValueError):
                print('Warning: key "' + subsection + '/' + value +
                      '" not found in the Almanac response.')

        return almdata


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class SkyModel:
    """SkyModel for querying the Advanced SkyModel"""
    """Note: skycalc_cli v.1.4 output wavelength unit is nm (it is μm in v.1.3)"""

    def __init__(self):

        print("Note: skycalc_cli v.1.4 output wavelength unit is nm (it is μm in v.1.3)")
        self.stop_on_errors_and_exceptions = True
        self.data = None
        self.server = 'https://etimecalret-002.eso.org'
        self.url = self.server + '/observing/etc/api/skycalc'
        self.deleter_script_url = self.server + '/observing/etc/api/rmtmp'
        self.bugreport_text = ''
        self.tmpdir = ''
        self.params = {
            # Airmass. Alt and airmass are coupled through the plane parallel
            # approximation airmass=sec(z), z being the zenith distance
            # z=90°−Alt
            'airmass': 1.0,    # float range [1.0,3.0]

            # Season and Period of Night
            'pwv_mode': 'pwv',  # string  grid ['pwv','season']
            # integer grid [0,1,2,3,4,5,6] (0=all year, 1=dec/jan,2=feb/mar...)
            'season': 0,
            # third of night integer grid [0,1,2,3] (0=all year, 1,2,3 = third
            # of night)
            'time': 0,

            # Precipitable Water Vapor PWV
            # mm float grid [-1.0,0.5,1.0,1.5,2.5,3.5,5.0,7.5,10.0,20.0]
            'pwv': 3.5,

            # Monthly Averaged Solar Flux
            'msolflux': 130.0,  # s.f.u float > 0

            # Scattered Moon Light
            # Moon coordinate constraints: |z – zmoon| ≤ ρ ≤ |z + zmoon| where
            # ρ=moon/target separation, z=90°−target altitude and
            # zmoon=90°−moon altitude.
            # string grid ['Y','N'] flag for inclusion of scattered moonlight.
            'incl_moon': 'Y',
            # degrees float range [0.0,360.0] Separation of Sun and Moon as
            # seen from Earth ("moon phase")
            'moon_sun_sep': 90.0,
            # degrees float range [0.0,180.0] Moon-Target Separation ( ρ )
            'moon_target_sep': 45.0,
            # degrees float range [-90.0,90.0] Moon Altitude over Horizon
            'moon_alt': 45.0,
            # float range [0.91,1.08] Moon-Earth Distance (mean=1)
            'moon_earth_dist': 1.0,

            # Starlight
            # string  grid ['Y','N'] flag for inclusion of scattered starlight
            'incl_starlight': 'Y',

            # Zodiacal light
            # string grid ['Y','N'] flag for inclusion of zodiacal light
            'incl_zodiacal': 'Y',
            # degrees float range [-180.0,180.0] Heliocentric ecliptic
            # longitude
            'ecl_lon': 135.0,
            # degrees float range [-90.0,90.0] Ecliptic latitude
            'ecl_lat': 90.0,

            # Molecular Emission of Lower Atmosphere
            # string grid ['Y','N'] flag for inclusion of lower atmosphere
            'incl_loweratm': 'Y',
            # Emission Lines of Upper Atmosphere
            # string grid ['Y','N'] flag for inclusion of upper stmosphere
            'incl_upperatm': 'Y',
            # Airglow Continuum (Residual Continuum)
            # string grid ['Y','N'] flag for inclusion of airglow
            'incl_airglow': 'Y',

            # Instrumental Thermal Emission This radiance component represents
            # an instrumental effect. The emission is provided relative to the
            # other model components. To obtain the correct absolute flux, an
            # instrumental response curve must be applied to the resulting
            # model spectrum See section 6.2.4 in the documentation
            # http://localhost/observing/etc/doc/skycalc/
            # The_Cerro_Paranal_Advanced_Sky_Model.pdf
            # string grid ['Y','N'] flag for inclusion of instrumental thermal
            # radiation
            'incl_therm': 'N',
            'therm_t1': 0.0,   # K float > 0
            'therm_e1': 0.0,   # float range [0,1]
            'therm_t2': 0.0,   # K float > 0
            'therm_e2': 0.0,   # float range [0,1]
            'therm_t3': 0.0,   # float > 0
            'therm_e3': 0.0,   # K float range [0,1]

            # Wavelength Grid
            'vacair': 'vac',  # vac or air
            'wmin': 300.0,  # nm float range [300.0,30000.0] < wmax
            'wmax': 2000.0,  # nm float range [300.0,30000.0] > wmin
            # string grid ['fixed_spectral_resolution','fixed_wavelength_step',
            #              'user']
            'wgrid_mode': 'fixed_wavelength_step',
            # nm/step float range [0,30000.0] wavelength sampling step dlam
            # (not the res.element)
            'wdelta': 0.1,
            # float range [0,1.0e6] RESOLUTION is misleading, it is rather
            # lam/dlam where dlam is wavelength step (not the res.element)
            'wres': 20000,
            'wgrid_user': [500.0, 510.0, 520.0, 530.0, 540.0, 550.0],
            # convolve by Line Spread Function
            'lsf_type': 'none',   # string grid ['none','Gaussian','Boxcar']
            'lsf_gauss_fwhm': 5.0,      # wavelength bins float > 0
            'lsf_boxcar_fwhm': 5.0,     # wavelength bins float > 0
            'observatory': 'paranal',   # paranal
            # compute temperature and bolometric radiance
            'temp_flag': 1
        }

    def handle_exception(self, e, msg):
        print(msg)
        print(e)
        print(self.bugreport_text)
        if(self.stop_on_errors_and_exceptions):
            sys.exit()

    # handle the kind of errors we issue ourselves.
    def handle_error(self, msg, stop=True):
        print(msg)
        print(self.bugreport_text)
        if(self.stop_on_errors_and_exceptions):
            sys.exit()

    def retrieve_data(self, url):
        try:
            response = requests.get(url, stream=True)
            self.data = response.content
        except requests.exceptions.RequestException as e:
            self.handle_exception(
                e, 'Exception raised trying to get FITS data from ' + url)

    def write(self, local_filename):
        try:
            with open(local_filename, 'wb') as f:
                f.write(self.data)
        except IOError as e:
            self.handle_exception(
                e, 'Exception raised trying to write fits file ')

    def getdata(self):
        return self.data

    def delete_server_tmpdir(self, tmpdir):
        try:
            response = requests.get(self.deleter_script_url + '?d=' + tmpdir)
            deleter_response = response.text.strip()
            if(deleter_response != 'ok'):
                self.handle_error('Could not delete server tmpdir ' + tmpdir)
        except requests.exceptions.RequestException as e:
            pass # ignore the exception

    def call(self, test=False):
        # print 'self.url=',self.url
        # print 'self.params=',self.params
        try:
            response = requests.post(self.url, data=json.dumps(self.params))
        except requests.exceptions.RequestException as e:
            self.handle_exception(
                e, 'Exception raised trying to POST request ' + self.url)
        try:
            res = json.loads(response.text)
            status = res['status']
            tmpdir = res['tmpdir']
        except (KeyError, ValueError) as e:
            self.handle_exception(
                e, 'Exception raised trying to decode server response ')

        tmpurl = self.server + '/observing/etc/tmp/' + \
            tmpdir + '/skytable.fits'

        if(status == 'success'):
            try:
                # retrive and save FITS data (in memory)
                self.retrieve_data(tmpurl)
            except requests.exceptions.RequestException as e:
                self.handle_exception(
                    e, 'could not retrieve FITS data from server')

            self.delete_server_tmpdir(tmpdir)

        else:  # print why validation failed
            self.handle_error('parameter validation error: ' +
                              res['error'])

        if(test):
            # print 'call() returning status:',status
            return status

    def callwith(self, newparams):
        for key, val in newparams.items():
            self.params[key] = val
        self.call()

    def printparams(self, keys=None):
        if keys is None:
            p = self.params
        else:
            p = dict((k, self.params[k]) for k in keys)
        for k in p:
            print(k, p[k])

    def test(self, label, overwite_params):
        from cStringIO import StringIO

        # capture stdout for a while
        old_stdout = sys.stdout
        sys.stdout = mystdout = StringIO()

        sm = SkyModel()
        str_params = ''
        # replace all matching keywords with the values in the newdict
        for key, val in overwite_params.items():
            if key in sm.params:
                sm.params[key] = val
                str_params = str_params + str(key) + ' : ' + str(val) + ', '
        status = sm.call(True)  # set True for test

        # restore original stdout
        sys.stdout = old_stdout
        toprint = label + '\t'
        # print str_params
        if(status == 'success'):
            toprint += bcolors.OKGREEN + ' *** pass *** ' + bcolors.ENDC
        else:
            toprint += bcolors.FAIL + ' *** fail *** ' + bcolors.ENDC
        print(toprint)

        print(mystdout.getvalue())

    def reset(self):
        self.__init__()



import numpy as np
import astropy.units as u


class TheoryModel:
    def __init__(self):
        self._flux = np.array([])
        self._wave = np.array([])

    @property
    def wave(self):
        return self._wave

    @wave.setter
    def wave(self,wave):
        self._lambmin = np.min(wave)
        self._lambmax = np.max(wave)
        self._wave = wave

    @property
    def lambmin(self):
        return self._lambmin

    @property
    def lambmax(self):
        return self._lambmax

    @property
    def x(self):
        return np.log(self._wave/u.Angstrom)

    def plot(self,ax,epoch_idx,normalize=None,nargs=[]):
        import matplotlib.pyplot as plt
        y = self.flux[epoch_idx,:]
        if normalize is not None:
            y = normalize(y,*nargs)
        ax.plot(self.x[epoch_idx,:], y,'o',color=self.color,alpha=0.4,label='Truth ' + self.__class__.__name__,markersize=4)
        return ax

    def plot_interpolated(self,ax,epoch_idx,normalize=None,nargs=[]):
        import matplotlib.pyplot as plt
        # import matplotlib.pyplot as plt
        y = self.fs[epoch_idx,:]
        if normalize is not None:
            y = normalize(y,*nargs)
        ax.plot(self.xs, y,'.',color=self.color,alpha=0.3,label='Interpolated ' + self.__class__.__name__,markersize=3)
        return ax



from simulacra.theory import TheoryModel
import scipy.io
import astropy.units as u
import numpy as np

def read_in_idl(filename='data/gascell/keck_fts_renorm.idl'):
    array = scipy.io.readsav(filename)
    transmission = array['siod']
    wavelength   = array['wiod']
    return transmission, wavelength

class GasCellModel(TheoryModel):
    def __init__(self,filename='../data/gascell/keck_fts_renorm.idl'):
        self._name = 'gascell'
        self.filename = filename
        transmission, wavelength = read_in_idl(self.filename)
        self.flux = transmission
        self.wave = wavelength * u.Angstrom

    def generate_transmission(self,star,detector,obs_times,exp_times):
        flux = [self.flux for i in range(len(obs_times))]
        wave = [self.wave for i in range(len(obs_times))]
        return flux, wave



from simulacra.theory import TheoryModel
import numpy.random as random
import astropy.units as u
import astropy.coordinates as coord
import astropy.time as at
import telfit
import logging
import numpy as np

def get_skycalc_defaults(inputFilename,almFilename,isVerbose=False):
    dic = {}

    # Query the Almanac if alm option is enabled
    if almFilename:

        # Read the input parameters
        inputalmdic = None
        try:
            with open(almFilename, 'r') as f:
                inputalmdic = json.load(f)
        except ValueError:
            with open(almFilename, 'r') as f:
                inputalmdic = sky_cli.loadTxt(f)

        if not inputalmdic:
            raise ValueError('Error: cannot read' + almFilename)

        alm = skycalc.AlmanacQuery(inputalmdic)
        dic = alm.query()

    if isVerbose:
        print('Data retrieved from the Almanac:')
        for key, value in dic.items():
            print('\t' + str(key) + ': ' + str(value))

    if inputFilename:

        # Read input parameters
        inputdic = None
        try:
            with open(inputFilename, 'r') as f:
                inputdic = json.load(f)
        except ValueError:
            with open(inputFilename, 'r') as f:
                inputdic = sky_cli.loadTxt(f)

        if not inputdic:
            raise ValueError('Error: cannot read ' + inputFilename)

        # Override input parameters
        if isVerbose:
            print('Data overridden by the user\'s input file:')
        for key, value in inputdic.items():
            if isVerbose and key in dic:
                print('\t' + str(key) + ': ' + str(value))
            dic[key] = value

    # Fix the observatory to fit the backend
    try:
        dic = sky_cli.fixObservatory(dic)
    except ValueError:
        raise

    if isVerbose:
        print('Data submitted to SkyCalc:')
        for key, value in dic.items():
            print('\t' + str(key) + ': ' + str(value))

    return dic

def sample_airmass(epoches):
    airmass = (2 * random.rand(epoches)) + 1
    return airmass

class TelluricsModel(TheoryModel):
    def __init__(self):
        pass
        # self._airmass = np.array([])

    # def generate_airmass(self,epoches,airmassmin=1.0,airmassmax=3.0):
    #     self._airmass = sample_airmass(epoches)

    # @property
    # def airmass(self,airmass):
        return self._airmass


class SkyCalcModel(TelluricsModel):
    def __init__(inputfile,almfile):
        super(SkyCalcModel,self).__init__()
        self.inputfile = inputfile
        self.almfile   = almfile

    def generate_transmission(epoches):
        # airmass = sample_airmass(epoches)
        dic     = get_skycalc_defaults(self.inputfile,self.almfile)
        lambda_grid       = []
        transmission_grid = []
        for a in self.airmass:
            dic['airmass'] = a
            skyModel = skycalc.SkyModel()
            skyModel.callwith(dic)

            data = skyModel.getdata()
            data = io.BytesIO(data)
            hdu = at.Table.read(data)
            transmission_grid.append(hdu['trans'].data)
            lambda_grid.append(hdu['lam'].data)
        self.flux = np.array(transmission_grid)
        self.wave = np.array(lambda_grid) * u.nm
        return np.array(transmission_grid), np.array(lambda_grid) * u.nm


class TelFitModel(TelluricsModel):
    def __init__(self,lambmin,lambmax,loc,humidity=50.0,temperature=300*u.Kelvin,pressure=1.0e6*u.Pa,wave_padding=10*u.Angstrom):
        self._name = 'tellurics'
        self.wave_padding = wave_padding
        self.lambmin = lambmin
        self.lambmax = lambmax
        # assert that all of these parameters that are arrays
        # are the same length
        # or they are just a scalar
        self.epoches     = None
        self.temperature = temperature
        self.pressure    = pressure
        self.humidity    = humidity

        # dlamb = 1e-2 * u.Angstrom
        # self.wave = np.arange(lambmin.to(u.Angstrom).value,lambmax.to(u.Angstrom).value,step=dlamb.to(u.Angstrom).value) * u.Angstrom

        self.loc = loc

    def check_shape_type(self,value,unit=None):
        if isinstance(value, u.Quantity):
            if value.unit.physical_type in unit.physical_type:
                if self.epoches:
                    if value.shape[0] == self.epoches:
                        return value
                    elif len(value.shape) == 0:
                        return value * np.ones(self.epoches)
                    else:
                        logging.warn('input not the correct size {}'.format(self.epoches))
                else:
                    if len(value.shape) == 0:
                        return value
                    else:
                        self.epoches = value.shape[0]
                        return value
        logging.warn('type or physical type are not correct')
        return None

    def temperature():
        doc = "The temperature property."
        def fget(self):
            return self._temperature
        def fset(self, value):
            out = self.check_shape_type(value,unit=u.Kelvin)
            if out is not None:
                self._temperature = out
            else:
                print('temperature left as default {}'.format(self.temperature))

        def fdel(self):
            del self._temperature
        return locals()
    temperature = property(**temperature())

    def pressure():
        doc = "The pressure property."
        def fget(self):
            return self._pressure
        def fset(self,value):
            out = self.check_shape_type(value,unit=u.Pa)
            if out is not None:
                self._pressure = out
            else:
                print('pressure left as default {}'.format(self.pressure))

        def fdel(self):
            del self._pressure
        return locals()
    pressure = property(**pressure())

    def humidity():
        doc = "The humidity property."
        def fget(self):
            return self._humidity
        def fset(self, value):
            if isinstance(value, float):
                self._humidity = value * np.ones(self.epoches)
            elif isinstance(value, np.ndarray):
                if value.shape[0] == self.epoches:
                    self._humidity = value
                else:
                    logging.warn('ndarray incorrect shape')
            else:
                logging.warn('incorrect type: must be float or ndarray')

        def fdel(self):
            del self._humidity
        return locals()
    humidity = property(**humidity())

    def check_property(self,property,value):
        if len(property.shape) == 0:
            return property * np.ones(value)
        elif property.shape[0] < value:
            logging.warn('epoches larger than the set temperatures\npadding with last element')
            temp = property
            return np.concatenate((temp,temp[-1] * np.ones(value - temp.shape[0])))
        elif property.shape[0] > value:
            logging.warn('epoches smaller than the set temperatures\ntruncating elements')
            return property[:value]
        else:
            return property

    def epoches():
        doc = "The epoches property."
        def fget(self):
            return self._epoches
        def fset(self,value):
            self._epoches = value
            try:
                self.temperature = self.check_property(self.temperature,value)
            except AttributeError:
                pass
            try:
                self.pressure = self.check_property(self.pressure,value)
            except AttributeError:
                pass
            try:
                self.humidity = self.check_property(self.humidity,value)
            except AttributeError:
                pass
        def fdel(self):
            del self._epoches
        return locals()
    epoches = property(**epoches())

    def generate_transmission(self,star,detector,obs_times,exp_times):
        times = at.Time([obs_times[i] + exp_times[i]/2 for i in range(len(obs_times))])
        telescope_frame = coord.AltAz(obstime=obs_times,location=detector.loc)
        secz = np.array(star.target.transform_to(telescope_frame).secz)
        if self.epoches != obs_times.shape[0]:
            logging.warning('tellurics epoches not the same as obs times\nresetting...')
            self.epoches = obs_times.shape[0]

        modeler = telfit.Modeler(debug=True)
        flux = []
        wave = []
        # print(self.lambmin.to(u.cm),self.lambmax.to(u.cm))
        for i,time in enumerate(obs_times):

            angle = np.arccos(1./secz[i]) * 180 * u.deg/np.pi
            print('humidity: {}\n'.format(self.humidity[i]),
                'pressure: {}\n'.format(self.pressure[i].to(u.hPa).value),
                'temperature: {}\n'.format(self.temperature[i].to(u.Kelvin).value),
                'lat: {}\n'.format(self.loc.lat.to(u.degree).value),
                'elevation: {}\n'.format(self.loc.height.to(u.km).value),
                'freqmin(cm-1): {}\n'.format(1.0/(self.lambmax.to(u.cm).value)),
                'freqmax(cm-1): {}\n'.format(1.0/(self.lambmin.to(u.cm).value)),
                'angle: {}\n'.format(angle.to(u.deg).value))

            model = modeler.MakeModel(humidity=self.humidity[i],
                         pressure=float(self.pressure[i].to(u.hPa).value),
                         temperature=float(self.temperature[i].to(u.Kelvin).value),
                         lat=float(self.loc.lat.to(u.degree).value),
                         alt=float(self.loc.height.to(u.km).value),
                         lowfreq=float(1.0/self.lambmax.to(u.cm).value),
                         highfreq=float(1.0/self.lambmin.to(u.cm).value),
                         angle=float(angle.to(u.deg).value))

            ns   = len(model.x)
            print(ns)
            flux.append(model.y)
            wave.append(model.x * u.nm)

        # flux = flux.reshape(self.epoches,ns)
        # wave = wave.reshape(self.epoches,ns)
        return flux, wave

    @TheoryModel.lambmin.setter
    def lambmin(self,lambmin):
        self._lambmin = lambmin - self.wave_padding

    @TheoryModel.lambmax.setter
    def lambmax(self,lambmax):
        self._lambmax = lambmax + self.wave_padding



import numpy as np

import astropy.units as u
import astropy.constants as const
import astropy.io.fits
import astropy.time as atime
import astropy.coordinates as coord

import numpy.random as random
import os

from simulacra.theory import TheoryModel

def sample_deltas(epoches,vel_width=30*u.km/u.s):
    deltas  = np.array(shifts((2*random.rand(epoches)-1)*vel_width))
    return deltas

def read_in_fits(filename):
    print('reading in {}'.format(filename))
    grid = astropy.io.fits.open(filename)['PRIMARY'].data
    # flux_all    = astropy.io.fits.open(fluxfile)['PRIMARY'].data
    return grid

def download_phoenix_wave(outdir):
    filename = 'WAVE_PHOENIX-ACES-AGSS-COND-2011.fits'
    outname = os.path.join(outdir,filename)
    if os.path.isfile(outname):
        print('using saved wave file')
        return outname
    else:
        from ftplib import FTP

        ftp = FTP('phoenix.astro.physik.uni-goettingen.de') #logs in
        ftp.login()
        ftp.cwd('HiResFITS')
        ftp.retrlines('LIST')

        with open(outname, 'wb') as fp:
            ftp.retrbinary('RETR ' + filename, fp.write) # start downloading

        ftp.close() # close the connection

        return outname

def download_phoenix_model(star,outdir=None):
    directories = ['data','HiResFITS','PHOENIX-ACES-AGSS-COND-2011','Z{:+.1f}'.format(star.z)]
    # print(directories)
    if star.alpha != 0.0:
        directories[-1] += '.Alpha={:+.2f}'.format(star.alpha)

    filename = 'lte{:05d}-{:1.2f}'.format(star.temperature,star.logg) + directories[-1][1:] + '.PHOENIX-ACES-AGSS-COND-2011-HiRes.fits'
    outname = os.path.join(outdir,filename)
    print(outname)
    if os.path.isfile(outname):
        print('using saved flux file')
        return outname

    else:
        from ftplib import FTP

        ftp = FTP('phoenix.astro.physik.uni-goettingen.de') #logs in
        ftp.login()
    #     for directory in directories:
        print("ftp: {}\nfilename: {}\ndirs: {}".format(ftp, filename,directories))
        ftp.cwd(os.path.join(*directories))
        ftp.retrlines('LIST')


        with open(outname, 'wb') as fp:
            ftp.retrbinary('RETR ' + filename, fp.write) # start downloading

        ftp.close() # close the connection

        return outname

def zplusone(vel):
    return np.sqrt((1 + vel/(const.c))/(1 - vel/(const.c)))

def shifts(vel):
    return np.log(zplusone(vel))

def get_random_times(n,tframe=365*u.day):
    now = atime.Time.now()
    dts = np.random.uniform(0,tframe.value,n) * tframe.unit
    times = now + dts
    return times

def get_berv(times,obs,ra,dec,v_d):
    obj = coord.SkyCoord(ra,dec,radial_velocity=v_d)
    loc = coord.EarthLocation.of_site(obs)
    bc  = obj.radial_velocity_correction(obstime=times,location=loc).to(u.km/u.s)
    return bc

def binary_system_velocity(times,amplitude,period,phase_time='2000-01-02'):
    starttime = atime.Time(phase_time)
    ptime = (times - starttime)/period
    return amplitude * np.sin(2*np.pi*ptime*u.radian)

def get_velocity_measurements(times,amplitude,period,loc,target):
    # berv  = get_berv(times,obs,ra,dec,velocity_drift)
    berv = target.radial_velocity_correction(obstime=times,location=loc)
    rvs  = berv + binary_system_velocity(times,amplitude,period)
    return rvs

def get_night_grid(loc,tstart,tend,steps_per_night=10):
    from astroplan import Observer
    observer = Observer(location=loc)
    days = int((tend - tstart) / u.day)

    # print(days)
    all_times = tstart + np.linspace(0,days,days + 1)
    sunset  = observer.sun_set_time(all_times,which='next')
    sunrise = observer.sun_rise_time(all_times,which='next')

    # print(all_times)
    outarr = np.array([])
    for i in range(len(sunrise)-1):
        nighttime = (sunrise[i+1]-sunset[i])
        outarr = np.concatenate((outarr,
                                sunset[i] + \
                                np.linspace(0,nighttime.value,steps_per_night)))
    return outarr

def get_realistic_times(target,loc,all_times):

    telescope_frame = coord.AltAz(obstime=all_times,location=loc)
    secz = np.array(target.transform_to(telescope_frame).secz)
    isvis = secz > 0

    # time? alltimes
    possible_times = all_times[isvis]
    airmass = secz[isvis]

    isreasonable = airmass < 3.0
    return possible_times[isreasonable], airmass[isreasonable]


class StarModel(TheoryModel):
    def __init__(self,deltas):
        # super(StarModel,self).__init__()
        self._deltas = deltas

    @property
    def deltas(self):
        return self._deltas

    @deltas.setter
    def deltas(self,deltas):
        self._deltas = deltas

def stellar_to_detector_flux(star,detector,exp_times):
    stellar_area = 4. * np.pi * star.stellar_radius**2
    ratio_of_areas = detector.area / (4.* np.pi * star.distance**2)

    print("photon flux: {:.2e}".format(np.mean(star.wave_difference * star.surface_flux.to(u.photon/u.s / u.m**3, u.spectral_density(star.wave))).to(u.ph/u.m**2/u.s).value))
    print("ratios: {:.2e}".format(ratio_of_areas.to(1).value))
    print("exposures: {:.2e}".format(exp_times[0].to(u.s).value))
    print("star area: {:.2e}".format(stellar_area.to(u.m**2).value))
    det_flux = detector.through_put * np.outer(exp_times, np.multiply(star.surface_flux.to(u.photon/u.s / u.m**3, \
                            u.spectral_density(star.wave)) \
                            , star.wave_difference)) * stellar_area * ratio_of_areas
    print("det flux: {:.2e}".format(np.mean(det_flux).to(u.ph).value))
    return det_flux.to(u.ph)

class PhoenixModel(TheoryModel):
    def __init__(self,distance,alpha,z,temperature,logg,target,amplitude,period,outdir=None):
        super(PhoenixModel,self).__init__()
        if outdir is None:
            self.outdir = os.path.join('..','data','stellar','PHOENIX')
            os.makedirs(self.outdir,exist_ok=True)
        self.temperature = temperature
        self.z = z
        self.logg  = logg
        self.alpha = alpha
        self.wavename = download_phoenix_wave(self.outdir)
        self.fluxname = download_phoenix_model(self,self.outdir)

        grid = astropy.io.fits.open(self.fluxname)
        self.stellar_radius = grid['PRIMARY'].header['PHXREFF'] * u.cm
        self.surface_flux = grid['PRIMARY'].data * u.erg / u.cm**3 / u.s
        self.wave     = read_in_fits(self.wavename) * u.Angstrom

        # make these attributes of the phoenix model
        self.distance = coord.Distance(distance)
        self.target = target
        self.amplitude = amplitude
        self.period    = period

    def wave_difference():
        doc = "The wave_difference property."
        def fget(self):
            diff = 0.1 * u.Angstrom * np.ones(self.wave.shape)
            return diff
        return locals()
    wave_difference = property(**wave_difference())

    def generate_spectra(self,detector,obs_times,exp_times):
        # add integral over transmission
        time = atime.Time([obs_times[i] + exp_times[i]/2 for i in range(len(obs_times))])
        rvs    = get_velocity_measurements(time,self.amplitude,self.period,detector.loc,self.target)
        deltas = shifts(rvs)

        print('surface flux: mean {:3.2e}\t median {:3.2e}'.format(np.mean(self.surface_flux),np.median(self.surface_flux)))
        obs_flux = self.surface_flux * (self.stellar_radius**2/self.distance**2).to(1)
        print('obs     flux: mean {:3.2e}\t median {:3.2e}'.format(np.mean(obs_flux),np.median(obs_flux)))
        # axes.plot(self.wave,obs_flux,'or',alpha=0.3)
        # # axes.set_xlim(6120,6130)
        # plt.show()
        obs_flux = np.outer(np.ones(obs_times.shape),obs_flux)
        # obs_flux = stellar_to_detector_flux(self,detector,exp_times)
        return obs_flux, self.wave, deltas, rvs

    def plot(self,ax,epoch_idx,normalize=None,nargs=[]):
        y = self.flux
        if normalize is not None:
            y = normalize(y,*nargs)
        ax.plot(self.x - self.deltas[epoch_idx],y,'o',color=self.color,alpha=0.4,label='Truth ' + self.__class__.__name__,markersize=4)
        return ax

    def plot_interpolated(self,ax,epoch_idx,normalize=None,nargs=[]):
        # import matplotlib.pyplot as plt
        y = self.fs[epoch_idx,:]
        if normalize is not None:
            y = normalize(y,*nargs)
        ax.plot(self.xs,y,'.',color=self.color,alpha=0.3,label='Interpolated ' + self.__class__.__name__,markersize=3)
        return ax



import numpy as np
import astropy.units as u

def lanczos_interpolation(x,xs,ys,dx,a=4):
    x0 = xs[0]
    y = np.zeros(x.shape)
    v_lanczos = np.vectorize(lanczos_kernel)
    for i,x_value in enumerate(x):
        # which is basically the same as sample=x[j-a+1] to x[j+a]
        # where j in this case is the nearest index xs_j to x_value
        sample_min,sample_max = max(0,abs(x_value-x0)//dx - a + 1),min(xs.shape[0],abs(x_value-x0)//dx + a)
        samples = np.arange(sample_min,sample_max,dtype=int)
        # y[i]/ = v_lanczos((x_value - xs[samples])/dx,a).sum()
        for sample in samples:
            y[i] += ys[sample] * lanczos_kernel((x_value - xs[sample])/dx,a)
    return y

def lanczos_kernel(x,a):
    if x == 0:
        return 1
    if x > -a and x < a:
        return a*np.sin(np.pi*u.radian*x) * np.sin(np.pi*u.radian*x/a)/(np.pi**2 * x**2)
    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# European Southern Observatory
# For any scientific or technical question,
# please contact usd-help@eso.org

from __future__ import print_function
import pkg_resources
import sys
import json
import getopt
# import skycalc #import SkyModel
# import skycalc import AlmanacQuery


def loadTxt(inputFile):

    result = {}

    lineN = 0
    for line in inputFile.readlines():

        lineN += 1

        originalLine = line
        line = line.strip()
        if len(line) == 0 or line[0] == '#':
            continue
        line = line.split(':')
        if len(line) > 2 and line[0].strip() == 'date':
            line = [line[0], ':'.join(line[1::])]
        elif len(line) != 2 or (len(line) == 2 and (line[0].strip() == '' or
                                                    line[1].strip() == '')):
            print('WARNING: input file line ' + str(lineN) +
                  ': wrong format: ignored.')
            print(originalLine.rstrip())
            continue
        key = line[0].strip()
        value = line[1].strip()

        # Is it an integer?
        try:
            value = int(value)
        except ValueError:
            # Is it an float?
            try:
                value = float(value)
            except ValueError:
                pass

        result[key] = value

    return result


def fixObservatory(indic):

    if 'observatory' in indic:
        if indic['observatory'] == 'lasilla':
            indic['observatory'] = '2400'
        elif indic['observatory'] == 'paranal':
            indic['observatory'] = '2640'
        elif (indic['observatory'] == '3060m' or
              indic['observatory'] == 'armazones'):
            indic['observatory'] = '3060'
#        elif indic['observatory'] == '5000m':
#            indic['observatory'] = '5000'
        else:
            raise ValueError('Wrong Observatory name, please refer to the '
                             'documentation.')
    return indic


def usage():

    print('usage: skycalc_cli -i|--in inputfile -o|--out file.fits ' +
          '[-a|--alm almanacparameterfile] [-v|--verbose] [--version]')

# print('hi')
# def main():
#     # print('hi')
#     argv = sys.argv[1:]
#
#     almFilename = None
#     inputFilename = None
#     outputFilename = None
#     isVerbose = False
#
#     # Read command line arguments
#     try:
#         opts, args = getopt.getopt(argv, "hva:i:o:", ["help", "verbose",
#                                                       "version",
#                                                       "alm=", "in=", "out="])
#     except getopt.GetoptError:
#         usage()
#         sys.exit(1)
#
#     for opt, arg in opts:
#         if opt in ("-h", "--help"):
#             usage()
#             sys.exit()
#         elif opt in ("--version"):
#             version = pkg_resources.require("skycalc_cli")[0].version
#             print('skycalc_cli version ' + version)
#             sys.exit(0)
#         elif opt in ("-v", "--verbose"):
#             isVerbose = True
#         elif opt in ("-a", "--alm"):
#             almFilename = arg
#         elif opt in ("-i", "--in"):
#             inputFilename = arg
#         elif opt in ("-o", "--out"):
#             outputFilename = arg
#
#     if not ((inputFilename or almFilename) and outputFilename):
#         usage()
#         sys.exit(1)
#
#     dic = {}
#
#     # Query the Almanac if alm option is enabled
#     if almFilename:
#
#         # Read the input parameters
#         inputalmdic = None
#         try:
#             with open(almFilename, 'r') as f:
#                 inputalmdic = json.load(f)
#         except ValueError:
#             with open(almFilename, 'r') as f:
#                 inputalmdic = loadTxt(f)
#
#         if not inputalmdic:
#             raise ValueError('Error: cannot read' + almFilename)
#
#         alm = skycalc.AlmanacQuery(inputalmdic)
#         dic = alm.query()
#
#     if isVerbose:
#         print('Data retrieved from the Almanac:')
#         for key, value in dic.items():
#             print('\t' + str(key) + ': ' + str(value))
#
#     if inputFilename:
#
#         # Read input parameters
#         inputdic = None
#         try:
#             with open(inputFilename, 'r') as f:
#                 inputdic = json.load(f)
#         except ValueError:
#             with open(inputFilename, 'r') as f:
#                 inputdic = loadTxt(f)
#
#         if not inputdic:
#             raise ValueError('Error: cannot read ' + inputFilename)
#
#         # Override input parameters
#         if isVerbose:
#             print('Data overridden by the user\'s input file:')
#         for key, value in inputdic.items():
#             if isVerbose and key in dic:
#                 print('\t' + str(key) + ': ' + str(value))
#             dic[key] = value
#
#     # Fix the observatory to fit the backend
#     try:
#         dic = fixObservatory(dic)
#     except ValueError:
#         raise
#
#     if isVerbose:
#         print('Data submitted to SkyCalc:')
#         for key, value in dic.items():
#             print('\t' + str(key) + ': ' + str(value))
#
#
#     # print
#     # Get the Sky
#     skyModel = skycalc.SkyModel()
#     skyModel.callwith(dic)
#     skyModel.write(outputFilename)
#
# if __name__ == '__main__':
#     main()
